'use strict';

const cColors = new Float32Array([
    0, 0, 0, 0, // 0:
    0, 0, 1, 1, // 1: (Z) high impedance

    0.2, 0.847, 0.1,  1, // 2: strong 0
    0.2, 0.847, 0.1,  1, // 3: strong 1
    0.9, 0.2, 0.2, 1, // 4: (x X) strong unknown

    .5, 1, 1, 1, // 5: vec
    1, 1, 0, 1, // 6: yellow
    1, 0, 1, 1, // 7: strange purple

    0, 1, 0, .5, // 8: (l L) weak 0
    0, 1, 1, .5, // 9: (h H) weak 1
    1, 0, 0, .5, // 10: (w W) weak unknown

    0, 0, 0, 0, // 11:
    0, 0, 0, 0, // 12:
    0, 0, 0, 0, // 13:
    0, 0, 0, 0, // 14:
    0, 0, 0, 0  // 15:
]);

const cTilts = new Float32Array([ // 14
    0, 0,  // 0
    1, -1, // 1
    1, 0,  // 2
    1, 1,  // 3
    -1, -1, // 4
    -1, 0,  // 5
    -1, 1   // 6
]);

const shaderer = (kind, src) => webgl2 => {
    const vShader = webgl2.createShader(webgl2[kind]);
    webgl2.shaderSource(vShader, src);
    webgl2.compileShader(vShader);
    return vShader;
};

const vertexShaderScalar = shaderer('VERTEX_SHADER', `#version 300 es
in uvec3 pos;
out vec4 v_color;
uniform vec2 scale;
uniform vec2 offset;
uniform vec4 colors[16];
uniform vec2 tilts[7];
uniform float tilt;
void main() {
  v_color = colors[pos.z];
  vec2 node = tilts[pos.y];
  gl_Position = vec4(
    float(pos.x) * scale.x + offset.x + node[1] * tilt,
    float(node[0]) * scale.y + offset.y,
    1, 1
  );
}
`);

const fragmentShader = shaderer('FRAGMENT_SHADER', `#version 300 es
precision mediump float;
in vec4 v_color;
out vec4 myOutputColor;
void main() {
  myOutputColor = v_color;
}
`);


const bar = [
    (f, t, a, b, c0, c1) => [].concat( // 0
        (f === 1) ? [a, 2, c1] : [],
        [a, 5, c0],
        [b, 5, c0],
        (t > 1) ? [b, 0, c0] : []
    ),
    (f, t, a, b, c0, c1) => [].concat( // 1
        (f === 0) ? [a, 5, c0] :
            (f > 1) ? [a, 0, c0] : [],
        [a, 2, c1],
        [b, 2, c1],
        (t > 1) ? [b, 0, c0] : []
    ),
    (f, t, a, b, c0, c1) => [a, 0, c0, b, 0, c1] // 2 = Z
];

const brick = [
    (f, t, a, b, c0, c1) => [].concat( // 0
        (f === 0) ? [a, 5, c0] :
            (f === 1) ? [a, 1, c1] : [],
        [a, 6, c0],
        (t === 0) ? [b, 5, c0] :
            [b, 4, c0],
        (t === 3) ? [b, 0, c0] : []
    ),
    (f, t, a, b, c0, c1) => [].concat( // 1
        (f === 0) ? [a, 4, c0, a, 3, c1] :
            (f === 1) ? [a, 2, c1] :
                (f === 2) ? [a, 3, c1] :
                    [a, 0, c0, a, 3, c1],
        (t === 1) ? [b, 3, c1] :
            [b, 1, c1],
        (t === 3) ? [b, 0, c0] : []
    ),
    (f, t, a, b, c0, c1) => [].concat( // 2
        (f === 0) ? [a, 4, 0] :
            (f === 1) ? [a, 1, 0] :
                [a, 0, 0],
        (t === 0) ? [b, 6, 0, b, 6, c0] :
            (t === 1) ? [b, 3, 0, b, 3, c1, b, 4, c0] :
                [b, 0, 0, b, 0, c0, b, 4, c0],
        (f === 0) ? [a, 4, c0, a, 3, c1] :
            (f === 1) ? [a, 6, c0, a, 1, c1] :
                [a, 6, c0, a, 0, c0, a, 3, c1],
        (t === 0) ? [b, 1, c1, b, 6, c0] :
            (t === 1) ? [b, 3, c1] :
                [b, 1, c1, b, 0, c0]
    )
];


class WebGL2WaveRender {
    /**
     * 
     * @param {{
     * grid: HTMLDivElement,
     * view: HTMLDivElement,
     * values: HTMLDivElement
     * }} elements
     * @param {{
     * chango: Record<string, {
     *      kind: string,
     *      wave: Array<number | string>,
     *      vao: WebGLVertexArrayObject
     * }>
     * view: Array<{ ref: string }>,
     * currentWires: Set<{
     *      kind: string,
     *      link: string,
     *      name: string,
     *      size: number,
     *      parent: object,
     *      type: string
     * }>,
     * time: number
     * }} globalLookup
     * @param {{
     * width: number,
     * height: number,
     * xScale: number,
     * xOffset: number,
     * yOffset: number,
     * yStep: number,
     * yDuty: number
     * }} pstate 
     * @param { Array } plugins
     */
    constructor(elements, globalLookup, pstate, plugins) {
        const canvas = document.createElement('canvas');
        elements.view.replaceChildren(canvas);

        this.elements = elements;
        this.globalLookup = globalLookup;
        this.canvas = canvas;
        this.pstate = pstate;
        this.plugins = plugins;

        const webgl2 = canvas.getContext('webgl2', {
            premultipliedAlpha: false,
            alpha: true,
            antialias: false,
            depth: false
        });
        this.webglLocation = this.initProgram(webgl2);
        this.verticesMap = this.makeVertex();
        this.initData();

        this.animationHandler = undefined;
    }

    /**
     * 
     * @param {WebGL2RenderingContext} webgl2 
     * @returns {{
     * colors: WebGLUniformLocation,
     * tilts: WebGLUniformLocation,
     * scale: WebGLUniformLocation,
     * offset: WebGLUniformLocation,
     * tilt: WebGLUniformLocation,
     * pos: number,
     * webgl2: WebGL2RenderingContext
     * }} 
    */
    initProgram(webgl2) {
        const program = webgl2.createProgram();
        webgl2.attachShader(program, vertexShaderScalar(webgl2));
        webgl2.attachShader(program, fragmentShader(webgl2));
        webgl2.linkProgram(program);
        webgl2.useProgram(program);
    
        const webglLocation = {
            colors: webgl2.getUniformLocation(program, 'colors'),
            tilts:  webgl2.getUniformLocation(program, 'tilts'),
            scale:  webgl2.getUniformLocation(program, 'scale'),
            offset: webgl2.getUniformLocation(program, 'offset'),
            tilt:   webgl2.getUniformLocation(program, 'tilt'),
            pos: webgl2.getAttribLocation(program, 'pos'),
            webgl2
        };
    
        return webglLocation;
    }

    /**
     * 
     * @returns {Map<string, Uint32Array>}
     */
    makeVertex() {
        const globalLookup = this.globalLookup;
        const time = globalLookup.time;
        const verticesMap = new Map();
        for (const id of Reflect.ownKeys(globalLookup.chango)) {
            const signalItem = globalLookup.chango[id];
            const { kind, wave } = signalItem;
            if (kind === 'bit') {
                const vertices = this.makeBitVertex(wave, time);
                verticesMap.set(id, vertices);
            } else if (kind === 'vec') {
                const vertices = this.makeVecVertex(wave, time);
                verticesMap.set(id, vertices);
            }
        }
        return verticesMap;
    }

    /**
     * 
     * @param {Array<string | number>} wave 
     * @param { number } time 
     * @returns {Uint32Array}
     */
    makeBitVertex(wave, time) {
        const vertices = [];
        const ilen = wave.length;
        for (let i = 0; i < ilen; i++) {
            const f = wave[(i === 0) ? 0 : (i - 1)];
            const [tim, val] = wave[i];
            const t = wave[(i === (ilen - 1)) ? i : (i + 1)];
            const tt = (i === (ilen - 1)) ? time : wave[i + 1][0];
            switch (val) {
                case 0: case 1: // 0 1
                    vertices.push(...bar[val](f[1], t[1], tim, tt, 2, 3));
                    break;
                case 2: case 3: // x X
                    vertices.push(...bar[2](f[1], t[1], tim, tt, 4, 4));
                    break;
                case 4: case 5: // z Z
                    vertices.push(...bar[2](f[1], t[1], tim, tt, 1, 1));
                    break;
                case 6: case 7: // u U uninitialized
                    vertices.push(...bar[2](f[1], t[1], tim, tt, 6, 6));
                    break;
                case 8: case 9: // w W weak unknown
                    vertices.push(...bar[2](f[1], t[1], tim, tt, 10, 10));
                    break;
                case 10: case 11: // l L
                    vertices.push(...bar[0](f[1], t[1], tim, tt, 8, 9));
                    break;
                case 12: case 13: // h H
                    vertices.push(...bar[1](f[1], t[1], tim, tt, 8, 9));
                    break;
                default:
                    vertices.push(...bar[2](f[1], t[1], tim, tt, 7, 7));
            }
        }

        return new Uint32Array(vertices);
    }

    /**
     * 
     * @param {Array<string | number>} wave 
     * @param { number } time 
     * @returns {Uint32Array}
     */
    makeVecVertex(wave, time) {
        const vertices = [];
        const ilen = wave.length;
        for (let i = 0; i < ilen; i++) {
            const [t1, val, msk] = wave[i];
            const t2 = (i === (ilen - 1)) ? time : wave[i + 1][0];

            if (msk) {
                vertices.push(...brick[2](2, 2, t1, t2, 4, 4)); // x,z?
            } else {
                vertices.push(
                    t1, 0, 0,
                    t2, 0, 0,
                    t2, 0, 5,
                    t2, 4, 5,
                    t1, 6, 5,
                    t1, 0, 5,
                    t1, 3, 5,
                    t2, 1, 5,
                    t2, 0, 5
                );
            }

        }

        return new Uint32Array(vertices);
    }

    initData() {
        const webglLocation = this.webglLocation;
        const webgl2 = webglLocation.webgl2;
        const globalLookup = this.globalLookup;
        const verticesMap = this.verticesMap;
        
        for (const id of Reflect.ownKeys(globalLookup.chango)) {
            const signalItem = globalLookup.chango[id];
            const vertices = verticesMap.get(id);
            
            const vertexBuffer = webgl2.createBuffer();
            webgl2.bindBuffer(webgl2.ARRAY_BUFFER, vertexBuffer);
            // 将初始化的顶点数据复制到 buffer 区域
            webgl2.bufferData(webgl2.ARRAY_BUFFER, vertices, webgl2.STATIC_DRAW);
    
            signalItem.vao = webgl2.createVertexArray();
            webgl2.bindVertexArray(signalItem.vao);
            webgl2.vertexAttribIPointer(webglLocation.pos, 3, webgl2.UNSIGNED_INT, 0, 0);
            webgl2.enableVertexAttribArray(webglLocation.pos);
            webgl2.uniform4fv(webglLocation.colors, cColors);
            webgl2.uniform2fv(webglLocation.tilts, cTilts);
        }
    }

    render() {
        if (this.animationHandler !== undefined) {
            cancelAnimationFrame(this.animationHandler);
        }
        
        const canvas = this.canvas;
        const webglLocation = this.webglLocation;
        const webgl2 = webglLocation.webgl2;
        const globalLookup = this.globalLookup;
        const verticesMap = this.verticesMap;
        const elements = this.elements;

        this.animationHandler = window.requestAnimationFrame(() => {
            const { width, height, xScale, xOffset, yOffset, yStep, yDuty } = this.pstate;
            const canvasHeight = height - 40;
            canvas.width = width;
            canvas.height = canvasHeight;
            
            // 设置 glsl 变量
            webgl2.uniform1f(webglLocation.tilt, 3 / width);
            webgl2.uniform2f(webglLocation.scale, 2 * xScale / width, yStep * yDuty / canvasHeight);
            // 设置 webgl 和 canvas 大小位置一致
            webgl2.viewport(0, 0, width, canvasHeight);
            // 清楚颜色缓冲区，也就是删除上一次的渲染结果
            webgl2.clear(webgl2.COLOR_BUFFER_BIT);

            // 根据 globalLookup 当前激活的需要渲染的信号进行渲染
            let index = 0;
            for (const signal of globalLookup.currentWires) {
                const signalItem = globalLookup.chango[signal.link];
                if (!signalItem) {
                    return;
                }

                webgl2.bindVertexArray(signalItem.vao);
                webgl2.uniform2f(webglLocation.offset, 
                    (2 * xOffset / width) - 1, 
                    (2 * yOffset - 2 * yStep * (index + .7)) / canvasHeight + 1
                );
                // 根据 vao 进行绘制
                const vertices = verticesMap.get(signal.link);
                // console.log(vertices);
                webgl2.drawArrays(webgl2.LINE_STRIP, 0, vertices.length / 3);

                index ++;
            }

            this.plugins.map(fn => fn(globalLookup, this.pstate, elements));
            this.animationHandler = undefined;
        });
    }
}

module.exports = WebGL2WaveRender;